@@ -1,7 +1,7 @@
 /*
  *  amtterm -- Intel AMT serial-over-lan client, gtk version.
  *
- *  Copyright (C) 2007 Gerd Hoffmann <kraxel@redhat.com
+ *  Copyright (C) 2007,08 Gerd Hoffmann <kraxel@redhat.com
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -54,6 +54,10 @@ struct gamt_window {
     struct redir   redir;
     GIOChannel     *ch;
     guint          id;
+
+    /* logging */
+    char           *logname;
+    FILE           *logfile;
 };
 
 static const char *state_stock[] = {
@@ -96,6 +100,82 @@ static void gamt_rebuild_hosts(struct gamt_window *gamt);
 
 /* ------------------------------------------------------------------ */
 
+static int log_getfile(struct gamt_window *gamt)
+{
+    GtkWidget *dialog;
+    int ret = -1;
+
+    dialog = gtk_file_chooser_dialog_new("Logfile",
+                                         GTK_WINDOW(gamt->win),
+                                         GTK_FILE_CHOOSER_ACTION_SAVE,
+                                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+                                         GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
+                                         NULL);
+
+    if (!gamt->logname) {
+        char defname[80];
+        snprintf(defname, sizeof(defname), "%s.log", gamt->redir.host);
+        gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(dialog), defname);
+    } else {
+        gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(dialog), gamt->logname);
+    }
+
+    if (gtk_dialog_run(GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
+        free(gamt->logname);
+        gamt->logname = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER (dialog));
+        ret = 0;
+    }
+    gtk_widget_destroy (dialog);
+    return ret;
+}
+
+static void log_start(struct gamt_window *gamt)
+{
+    GtkWidget *item;
+
+    if (gamt->logfile)
+        goto out;
+    if (!gamt->logname) {
+        if (log_getfile(gamt) < 0)
+            goto out;
+    }
+
+    gamt->logfile = fopen(gamt->logname, "a");
+    if (!gamt->logfile) {
+        fprintf(stderr, "warning: open log %s: %s\n",
+                gamt->logname, strerror(errno));
+        goto out;
+    }
+    setvbuf(gamt->logfile, NULL, _IONBF /* unbuffered */, 0);
+
+out:
+    item = gtk_ui_manager_get_widget(gamt->ui, "/MainMenu/TtyMenu/WriteLog");
+    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item),
+                                   gamt->logfile ? TRUE : FALSE);
+}
+
+static void log_write(struct gamt_window *gamt, unsigned char *buf, int len)
+{
+    if (!gamt->logfile)
+        return;
+    /* Hmm, filter out control codes? */
+    fwrite(buf, len, 1, gamt->logfile);
+}
+
+static void log_stop(struct gamt_window *gamt)
+{
+    GtkWidget *item;
+
+    if (!gamt->logfile)
+        return;
+    fclose(gamt->logfile);
+    gamt->logfile = NULL;
+    item = gtk_ui_manager_get_widget(gamt->ui, "/MainMenu/TtyMenu/WriteLog");
+    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), FALSE);
+}
+
+/* ------------------------------------------------------------------ */
+
 static void menu_cb_connect(GtkAction *action, void *data)
 {
     struct gamt_window *gamt = data;
@@ -119,7 +199,7 @@ static void menu_cb_connect(GtkAction *action, void *data)
 static void menu_cb_connect_to(GtkAction *action, void *data)
 {
     struct gamt_window *gamt = data;
-    
+
     if (gamt->redir.state != REDIR_NONE &&
 	gamt->redir.state != REDIR_CLOSED &&
 	gamt->redir.state != REDIR_ERROR)
@@ -140,11 +220,44 @@ static void menu_cb_disconnect(GtkAction *action, void *data)
     redir_sol_stop(&gamt->redir);
 }
 
+static void menu_cb_logfile(GtkAction *action, void *data)
+{
+    struct gamt_window *gamt = data;
+
+    if (log_getfile(gamt) < 0)
+        return;
+    if (gamt->logfile) {
+        log_stop(gamt);
+        log_start(gamt);
+    }
+}
+
+static void menu_cb_reset_terminal(GtkAction *action, void *data)
+{
+    struct gamt_window *gamt = data;
+
+    vte_terminal_reset(VTE_TERMINAL(gamt->vte), TRUE, TRUE);
+}
+
+static void menu_cb_write_log(GtkToggleAction *action, gpointer data)
+{
+    struct gamt_window *gamt = data;
+    gboolean state = gtk_toggle_action_get_active(action);
+
+    if (amt_debug)
+	fprintf(stderr, "%s: %s\n", __FUNCTION__, state ? "on" : "off");
+    if (state)
+        log_start(gamt);
+    else
+        log_stop(gamt);
+}
+
 static void menu_cb_config_font(GtkAction *action, void *data)
 {
     struct gamt_window *gamt = data;
     GtkWidget *dialog;
     char *fontname;
+    int done = 0;
 
     dialog = gtk_font_selection_dialog_new("Terminal font");
     fontname = cfg_get_str(CFG_FONT);
@@ -152,13 +265,21 @@ static void menu_cb_config_font(GtkAction *action, void *data)
 	(GTK_FONT_SELECTION_DIALOG(dialog), fontname);
 
     gtk_widget_show_all(dialog);
-    switch (gtk_dialog_run(GTK_DIALOG(dialog))) {
-    case GTK_RESPONSE_OK:
-	fontname = gtk_font_selection_dialog_get_font_name
-	    (GTK_FONT_SELECTION_DIALOG(dialog));
-	vte_terminal_set_font_from_string(VTE_TERMINAL(gamt->vte), fontname);
-	cfg_set_str(CFG_FONT, fontname);
-	break;
+    while (!done) {
+        switch (gtk_dialog_run(GTK_DIALOG(dialog))) {
+        case GTK_RESPONSE_OK:
+            done=1;
+            /* fall through */
+        case GTK_RESPONSE_APPLY:
+            fontname = gtk_font_selection_dialog_get_font_name
+                (GTK_FONT_SELECTION_DIALOG(dialog));
+            vte_terminal_set_font_from_string(VTE_TERMINAL(gamt->vte), fontname);
+            cfg_set_str(CFG_FONT, fontname);
+            break;
+        default:
+            done=1;
+            break;
+        }
     }
     gtk_widget_destroy(dialog);
 }
@@ -173,7 +294,7 @@ static int pickcolor(char *title, GdkColor *color)
     csel = GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(dialog)->colorsel);
     gtk_color_selection_set_has_opacity_control(csel, FALSE);
     gtk_color_selection_set_current_color(csel, color);
-    
+
     gtk_widget_show_all(dialog);
     switch (gtk_dialog_run(GTK_DIALOG(dialog))) {
     case GTK_RESPONSE_OK:
@@ -274,7 +395,7 @@ static void menu_cb_man_amt_howto(GtkAction *action, void *data)
 static void menu_cb_about(GtkAction *action, void *data)
 {
     static char *comments = "Intel AMT serial-over-lan client";
-    static char *copyright = "(c) 2007 Gerd Hoffmann";
+    static char *copyright = "(c) 2007,08 Gerd Hoffmann";
     static char *website = "http://dl.bytesex.org/releases/amtterm/";
     static char *authors[] = { "Gerd Hoffmann <kraxel@redhat.com>", NULL };
     struct gamt_window *gamt = data;
@@ -303,6 +424,8 @@ static void destroy_cb(GtkWidget *widget, gpointer data)
 static int recv_gtk(void *cb_data, unsigned char *buf, int len)
 {
     struct gamt_window *gamt = cb_data;
+
+    log_write(gamt, buf, len);
     vte_terminal_feed(VTE_TERMINAL(gamt->vte), buf, len);
     return 0;
 }
@@ -334,8 +457,8 @@ static void state_gtk(void *cb_data, enum redir_state old, enum redir_state new)
 	    gamt_rebuild_hosts(gamt);
 	/* fall through */
     default:
-	snprintf(buf, sizeof(buf), "%s: %s", gamt->redir.host,
-		 redir_state_desc(new));
+        snprintf(buf, sizeof(buf), "%s: %s", gamt->redir.host,
+                 redir_state_desc(new));
 	break;
     }
     if (state_stock[new])
@@ -359,12 +482,15 @@ static const GtkActionEntry entries[] = {
     {
 	.name        = "FileMenu",
 	.label       = "_File",
-    },{
-	.name        = "ConfigMenu",
-	.label       = "_Options",
     },{
 	.name        = "HostMenu",
 	.label       = "Ho_sts",
+    },{
+	.name        = "TtyMenu",
+	.label       = "_Terminal",
+    },{
+	.name        = "ConfigMenu",
+	.label       = "_Appearance",
     },{
 	.name        = "HelpMenu",
 	.label       = "_Help",
@@ -380,6 +506,11 @@ static const GtkActionEntry entries[] = {
 	.stock_id    = GTK_STOCK_DISCONNECT,
 	.label       = "_Disconnect",
 	.callback    = G_CALLBACK(menu_cb_disconnect),
+    },{
+	.name        = "Logfile",
+	.stock_id    = GTK_STOCK_SAVE,
+	.label       = "_Logfile ...",
+	.callback    = G_CALLBACK(menu_cb_logfile),
     },{
 	.name        = "Quit",
 	.stock_id    = GTK_STOCK_QUIT,
@@ -387,6 +518,12 @@ static const GtkActionEntry entries[] = {
 	.callback    = G_CALLBACK(menu_cb_quit),
     },{
 
+	/* Terminal menu */
+	.name        = "ResetTerminal",
+	.label       = "Reset",
+	.callback    = G_CALLBACK(menu_cb_reset_terminal),
+    },{
+
 	/* Config menu */
 	.name        = "VteFont",
 	.stock_id    = GTK_STOCK_SELECT_FONT,
@@ -423,6 +560,10 @@ static const GtkActionEntry entries[] = {
 
 static const GtkToggleActionEntry tentries[] = {
     {
+	.name        = "WriteLog",
+	.label       = "Enable logging",
+	.callback    = G_CALLBACK(menu_cb_write_log),
+    },{
 	.name        = "BlinkCursor",
 	.label       = "Blinking cursor",
 	.callback    = G_CALLBACK(menu_cb_blink_cursor),
@@ -436,10 +577,17 @@ static char ui_xml[] =
 "      <menuitem action='Connect'/>\n"
 "      <menuitem action='Disconnect'/>\n"
 "      <separator/>\n"
+"      <menuitem action='Logfile'/>\n"
+"      <separator/>\n"
 "      <menuitem action='Quit'/>\n"
 "    </menu>\n"
 "    <menu action='HostMenu'>\n"
 "    </menu>\n"
+"    <menu action='TtyMenu'>\n"
+"      <menuitem action='ResetTerminal'/>\n"
+"      <separator/>\n"
+"      <menuitem action='WriteLog'/>\n"
+"    </menu>\n"
 "    <menu action='ConfigMenu'>\n"
 "      <menuitem action='VteFont'/>\n"
 "      <menuitem action='VteForeground'/>\n"
@@ -479,7 +627,7 @@ static int gamt_getstring(GtkWidget *window, char *title, char *message,
     GtkWidget *dialog, *label, *entry;
     const char *txt;
     int retval;
-   
+
     /* Create the widgets */
     dialog = gtk_dialog_new_with_buttons(title,
 					 GTK_WINDOW(window),
@@ -490,7 +638,7 @@ static int gamt_getstring(GtkWidget *window, char *title, char *message,
 					 GTK_RESPONSE_REJECT,
                                          NULL);
     gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);
-    
+
     label = gtk_label_new(message);
     gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
 
@@ -578,7 +726,7 @@ static void gamt_rebuild_hosts(struct gamt_window *gamt)
 	gtk_action_group_add_actions(gamt->hosts_ag, &entry, 1, gamt);
     }
     pos += sprintf(hosts_xml+pos, "%s", hosts_xml_end);
-    
+
     /* add */
     gtk_ui_manager_insert_action_group(gamt->ui, gamt->hosts_ag, 1);
     gamt->hosts_id = gtk_ui_manager_add_ui_from_string(gamt->ui, hosts_xml, -1, &err);
@@ -591,7 +739,7 @@ static void gamt_rebuild_hosts(struct gamt_window *gamt)
 static int gamt_connect(struct gamt_window *gamt)
 {
     int rc;
-    
+
     if (0 == strlen(amt_pass)) {
 	char msg[128];
 
@@ -636,7 +784,7 @@ static struct gamt_window *gamt_window()
     gboolean state;
     struct gamt_window *gamt;
     char *str;
-    
+
     gamt = malloc(sizeof(*gamt));
     if (NULL == gamt)
 	return NULL;
@@ -684,14 +832,14 @@ static struct gamt_window *gamt_window()
     state = cfg_get_bool(CFG_BLINK, 0);
     gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), state);
     vte_terminal_set_cursor_blinks(VTE_TERMINAL(gamt->vte), state);
-    
+
     /* other widgets */
     gamt->status = gtk_label_new("idle");
     gtk_misc_set_alignment(GTK_MISC(gamt->status), 0, 0.5);
     gtk_misc_set_padding(GTK_MISC(gamt->status), 3, 1);
     gamt->icon = gtk_image_new_from_stock(GTK_STOCK_DISCONNECT,
 					  GTK_ICON_SIZE_SMALL_TOOLBAR);
-    
+
     /* Make a vbox and put stuff in */
     vbox = gtk_vbox_new(FALSE, 1);
     hbox = gtk_hbox_new(FALSE, 1);
@@ -716,7 +864,7 @@ static struct gamt_window *gamt_window()
 
     /* display window */
     gtk_widget_show_all(gamt->win);
-    
+
     str = cfg_get_str(CFG_FOREGROUND);
     if (str) {
 	gdk_color_parse(str, &color);
@@ -747,13 +895,14 @@ static void usage(FILE *fp)
             "   -p pass       password (default: $AMT_PASSWORD)\n"
             "   -f font       terminal font\n"
             "   -c color      text color\n"
-            "   -b color      backgrounf color\n"
+            "   -b color      background color\n"
+            "   -l file       logfile\n"
             "\n"
             "By default port 16994 is used.\n"
 	    "If no password is given " APPNAME " will ask for one.\n"
             "\n"
             "-- \n"
-            "(c) 2007 Gerd Hoffmann <kraxel@redhat.com>\n"
+            "(c) 2007,08 Gerd Hoffmann <kraxel@redhat.com>\n"
 	    "\n");
 }
 
@@ -763,7 +912,7 @@ main(int argc, char *argv[])
     Display *dpy;
     struct gamt_window *gamt;
     char configfile[256];
-    char *h;
+    char *h, *log = NULL;
     int c;
 
     if (NULL != (h = getenv("AMT_PASSWORD")))
@@ -784,7 +933,7 @@ main(int argc, char *argv[])
     fcntl(ConnectionNumber(dpy),F_SETFD,FD_CLOEXEC);
 
     for (;;) {
-        if (-1 == (c = getopt(argc, argv, "hdtu:p:f:c:b:")))
+        if (-1 == (c = getopt(argc, argv, "hdtu:p:f:c:b:l:")))
             break;
         switch (c) {
 	case 'd':
@@ -800,6 +949,10 @@ main(int argc, char *argv[])
 	    snprintf(amt_pass, sizeof(amt_pass), "%s", optarg);
 	    memset(optarg,'*',strlen(optarg)); /* rm passwd from ps list */
 	    break;
+	case 'l':
+            log = optarg;
+	    break;
+
 	case 'f':
 	    cfg_set_str(CFG_FONT, optarg);
 	    break;
@@ -823,11 +976,16 @@ main(int argc, char *argv[])
     if (NULL == gamt)
 	exit(1);
 
+    if (log) {
+        gamt->logname = strdup(log);
+        log_start(gamt);
+    }
+
     if (optind+1 <= argc) {
 	snprintf(amt_host, sizeof(amt_host), "%s", argv[optind]);
 	gamt_connect(gamt);
     }
-    
+
     gtk_main();
     cfg_write_file("config", configfile);
     exit(0);
